/* SZDD decruncher by Kyzer/CSG <kyzer@4u.net>
 * $VER: SZDD.c 1.1 (19.06.2000)
 * v1.1: better xfd name, eliminated 4k buffer
 *
 * based on "ms-expand" by Daniel Risacher <magnus@mit.edu> and someone else
 * http://www.im.lcs.mit.edu/~magnus/ms-expand.c
 *
 * Header:
 * 4 bytes : "SZDD"
 * 4 bytes : $88,$F0,$27,$33
 * 2 bytes : random value? (seen $41,$00 and $41,$39)
 * 4 bytes : A,B,C,D. unpacked size of file is (D<<24)|(C<<16)|(B<<8)|(A)
 * 
 * Compressed data format:
 * - one header byte, whose bits from LSB (0) to MSB (7) dictate the format
 *   of the next 8 elements
 * - if a bit is set, the next element is a single byte literal
 * - otherwise, the next element is a repeat: two bytes, %AAAAAAAA, %BBBBCCCC
 *   where offset = %BBBBAAAAAAAA and length is %CCCC + 3
 *
 * A virtual 4096 byte wraparound window must be kept by the decruncher.
 * A wraparound pointer starts at 4096-16 and each outputted byte must be
 * written at that position and the pointer advanced by one. The buffer
 * starts off filled with zeroes. The offsets of the compressed data refer
 * to an absolute position in this buffer, which may be before the pointer,
 * or after it.
 *
 * It seems that there can often be superfluous data at the end of a file -
 * I can only go by the uncompressed size, so I can only really test running
 * out of source data. Some files end in the middle of a run, so I can't
 * test if they exactly reach the endpoint and no more.
 */

#include <libraries/xfdmaster.h>
#include <proto/exec.h>
#include <exec/memory.h>
#include "SDI_compiler.h"

#define SZDD_OVERRUN  (144)    /* 8 matches, at maximum length */
#define SZDD_MINLEN   (14)     /* 4+4+2+4 = 14 byte header (and 0 data) */
#define SZDD_WNDSIZE  (4096)   /* LZ sliding window size (4k) */

static char version[]="$VER: SZDD 1.1 (19.06.2000) by <kyzer@4u.net>";


ASM(BOOL) SZDD_recog(REG(a0, STRPTR buf), REG(d0, ULONG length)
  REG(a1, struct xfdRecogResult *rr)) {
  if (buf[0]=='S' && buf[1]=='Z' && buf[2] =='D' && buf[3]=='D') {
    rr->xfdrr_MinTargetLen = SZDD_OVERRUN + (
      rr->xfdrr_FinalTargetLen =
      buf[10] | (buf[11]<<8) | (buf[12]<<16) | (buf[13]<<24)
    );
    return (BOOL) 1;
  }
  return (BOOL) 0;
}

ASM(BOOL) SZDD_decrunch(REG(a0, struct xfdBufferInfo * xbi),
  REG(a6, struct xfdMasterBase *xfdMasterBase)) {

  struct ExecBase *SysBase = xfdMasterBase->xfdm_ExecBase;

  UBYTE *rep, mask, bits;
  int offset, length, clrlen;

  UBYTE *src  = (UBYTE *) xbi->xfdbi_SourceBuffer + 14;
  UBYTE *dest = (UBYTE *) xbi->xfdbi_UserTargetBuf;
  UBYTE *ends = src  + xbi->xfdbi_SourceBufLen;
  UBYTE *endd = dest + xbi->xfdbi_TargetBufSaveLen;
  UBYTE *base = dest;
  int posn = SZDD_WNDSIZE-16;

  while (src < ends) {
    bits = *src++;
    for (mask = 0x01; mask & 0xFF; mask <<= 1) {
      if (bits & mask) posn++, *dest++ = *src++;
      else {
        offset  = *src++; length = *src++;
        offset |= (length << 4) & 0xF00;
        length  = (length & 0x0F) + 3;

        /* translate absolute 4k buffer offset into real window offset */
        posn &= SZDD_WNDSIZE-1; if (offset > posn) posn += SZDD_WNDSIZE;
        rep = dest - posn + offset;
        posn += length;

        /* if repeat starts before actual data, write zeroes */
        while (rep < base && length-- > 0) rep++, *dest++ = 0;

        /* copy out repeat */
        while (length-- > 0) *dest++ = *rep++;
      }

    }

    /* at worst, we could overrun dest by (15+3)*8 = 144 bytes */
    if (dest >= endd) return 1;
  }

  /* bad exit - run out of src */
  xbi->xfdbi_Error = XFDERR_CORRUPTEDDATA;
  return 0;
}

struct xfdSlave FirstSlave = {
  NULL, XFDS_VERSION, 39, "(SZDD) Microsoft Data Cruncher",
  XFDPFF_DATA|XFDPFF_RECOGLEN|XFDPFF_USERTARGET,
  0, (BOOL (*)()) SZDD_recog, (BOOL (*)()) SZDD_decrunch,
  NULL, NULL, 0, 0, SZDD_MINLEN
};
